@@ -53,6 +53,11 @@ def untested(meth):
     return meth
 
 class IloError(Exception):
+    def __init__(self, message, errorcode=None):
+        super(IloError, self).__init__(message)
+        self.errorcode = errorcode
+
+class IloCommunicationError(IloError):
     pass
 
 class IloLoginFailed(IloError):
@@ -201,7 +206,7 @@ class Ilo(object):
         except socket.sslerror: # Connection closed
             e = sys.exc_info()[1]
             if not data:
-                raise IloError("Communication with %s:%d failed: %s" % (self.hostname, self.port, str(e)))
+                raise IloCommunicationError("Communication with %s:%d failed: %s" % (self.hostname, self.port, str(e)))
 
         self._debug(1, "Received %d bytes" % len(data))
         self._debug(2, data)
@@ -216,7 +221,7 @@ class Ilo(object):
                 sp = subprocess.Popen([self.hponcfg, '--input', '--xmlverbose'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=None)
             except OSError:
                 e = sys.exc_info()[1]
-                raise IloError("Cannot run %s: %s" % (self.hponcfg, str(e)))
+                raise IloCommunicationError("Cannot run %s: %s" % (self.hponcfg, str(e)))
             sp.write = sp.stdin.write
             sp.read = sp.stdout.read
             return sp
@@ -227,15 +232,15 @@ class Ilo(object):
         try:
             sock.connect((self.hostname, self.port))
         except socket.timeout:
-            raise IloError("Timeout connecting to %s:%d" % (self.hostname, self.port))
+            raise IloCommunicationError("Timeout connecting to %s:%d" % (self.hostname, self.port))
         except socket.error:
             e = sys.exc_info()[1]
-            raise IloError("Error connecting to %s:%d: %s" % (self.hostname, self.port, str(e)))
+            raise IloCommunicationError("Error connecting to %s:%d: %s" % (self.hostname, self.port, str(e)))
         try:
             return ssl.wrap_socket(sock, ssl_version=ssl.PROTOCOL_TLSv1)
         except socket.sslerror:
             e = sys.exc_info()[1]
-            raise IloError("Cannot establish ssl session with %s:%d: %s" % (self.hostname, self.port, e.message or str(e)))
+            raise IloCommunicationError("Cannot establish ssl session with %s:%d: %s" % (self.hostname, self.port, e.message or str(e)))
 
     def _communicate(self, xml, protocol, progress=None):
         sock = self._get_socket()
@@ -303,7 +308,7 @@ class Ilo(object):
         except socket.sslerror: # Connection closed
             e = sys.exc_info()[1]
             if not data:
-                raise IloError("Communication with %s:%d failed: %s" % (self.hostname, self.port, str(e)))
+                raise IloCommunicationError("Communication with %s:%d failed: %s" % (self.hostname, self.port, str(e)))
 
         self._debug(1, "Received %d bytes" % len(data))
 
@@ -375,10 +380,12 @@ class Ilo(object):
                         # This is triggered when doing protocol detection, ignore
                         pass
                     else:
-                        if int(child.get('STATUS'), 16) in IloLoginFailed.possible_codes or \
-                                child.get('MESSAGE') in IloLoginFailed.possible_messages:
-                            raise IloLoginFailed
-                        raise IloError("Error communicating with iLO: %s" % child.get('MESSAGE'))
+                        status = int(child.get('STATUS'), 16)
+                        message = child.get('MESSAGE')
+                        if status in IloLoginFailed.possible_codes or \
+                                 message in IloLoginFailed.possible_messages:
+                            raise IloLoginFailed(message, status)
+                        raise IloError(message, status)
                 # And this type of message is the actual payload.
                 else:
                     return message
@@ -451,15 +458,26 @@ class Ilo(object):
         fd.close()
         return ret
 
-    def _info_tag(self, infotype, tagname, returntag=None, attrib={}):
+    def _info_tag(self, infotype, tagname, returntags=None, attrib={}):
         root, inner = self._root_element(infotype, MODE='read')
         etree.SubElement(inner, tagname, **attrib)
         header, message = self._request(root)
-        message = message.find(returntag or tagname)
-        if list(message):
-            return self._element_children_to_dict(message)
-        else:
-            return self._element_to_dict(message)
+
+        # Use tagname as returntags if returntags isn't specificed
+        if isinstance(returntags, basestring):
+            returntags = [returntags]
+        elif returntags is None:
+            returntags = [tagname]
+
+        for tag in returntags:
+            if message.find(tag) is None:
+                continue
+            message = message.find(tag)
+            if list(message):
+                return self._element_children_to_dict(message)
+            else:
+                return self._element_to_dict(message)
+        raise IloError("Expected tag '%s' not found" % "' or '".join(returntags))
 
     def _info_tag2(self, infotype, tagname, returntag=None, key=None):
         root, inner = self._root_element(infotype, MODE='read')
@@ -549,7 +567,7 @@ class Ilo(object):
         if computer_lock_key:
             computer_lock = "custom"
         if not computer_lock:
-            raise IloError("A value must be specified for computer_lock")
+            raise ValueError("A value must be specified for computer_lock")
         elements = [etree.Element('COMPUTER_LOCK', VALUE=computer_lock)]
         if computer_lock_key:
             elements.append(etree.Element('COMPUTER_LOCK_KEY', VALUE=computer_lock_key))
@@ -690,7 +708,11 @@ class Ilo(object):
 
     def get_one_time_boot(self):
         """Get the one time boot state of the host"""
-        return self._info_tag('SERVER_INFO', 'GET_ONE_TIME_BOOT', 'ONE_TIME_BOOT')
+        # Inconsistency between iLO 2 and 3, let's fix that
+        data = self._info_tag('SERVER_INFO', 'GET_ONE_TIME_BOOT', ('ONE_TIME_BOOT', 'GET_ONE_TIME_BOOT'))
+        if 'device' in data['boot_type']:
+            data['boot_type'] = data['boot_type']['device']
+        return data
 
     def get_persistent_boot(self):
         """Get the boot order of the host"""
@@ -1062,3 +1084,87 @@ class Ilo(object):
         """FIXME: I have no relevant hardware. Please report sample output"""
         return self._info_tag('RIB_INFO', 'GET_VF_STATUS')
 
+###############################################################################
+### Testsuite, not to be run by end-users
+### Note: use only assertTrue/assertRaises for compatibility with older python
+### versions
+    def _test(self):
+        import unittest
+        this_ilo = self
+        class IloTest(unittest.TestCase):
+            @classmethod
+            def setUpClass(self):
+                sys.stdout.write("Identifying iLO version... ")
+                sys.stdout.flush()
+                self.ilo = this_ilo
+                res = self.ilo.get_fw_version()
+                print(res['management_processor'])
+                self.ilo_version = int(res['management_processor'][3:] or 1)
+                print("Running tests. This will take a few minutes")
+
+            def test_get_all_users(self):
+                users = self.ilo.get_all_users()
+                self.assertTrue(isinstance(users, list))
+                self.assertTrue(self.ilo.login in users)
+
+            def test_get_embedded_health(self):
+                if self.ilo_version < 2:
+                    return
+                res = self.ilo.get_embedded_health()
+                self.assertTrue(isinstance(res, dict))
+                self.assertTrue('temperature' in res)
+
+            def test_get_cert_subject_info(self):
+                if self.ilo_version < 2:
+                    return
+                res = self.ilo.get_cert_subject_info()
+                self.assertTrue(isinstance(res, dict))
+
+            def test_get_host_data(self):
+                res = self.ilo.get_host_data()
+                res2 = self.ilo.get_host_data(decoded_only=False)
+                self.assertTrue(isinstance(res, list))
+                self.assertTrue(isinstance(res2, list))
+                self.assertTrue(len(res2) > len(res))
+                self.assertTrue(self.issubset(res, res2))
+                self.assertTrue('b64_data' in res[0])
+
+            def test_get_host_power_status(self):
+                res = self.ilo.get_host_power_status()
+                self.assertTrue(isinstance(res, dict))
+                self.assertTrue('host_power_saver' in res)
+
+            def test_get_one_time_boot(self):
+                if self.ilo_version < 2:
+                    return
+                res = self.ilo.get_one_time_boot()
+                self.assertTrue(isinstance(res, dict))
+                self.assertTrue('boot_type' in res)
+
+            def test_get_server_power_on_time(self):
+                res
+
+            # Tests to write:
+            #
+            # get_server_power_on_time
+            # get_sso_settings
+            # Auth error
+            # Connect error
+            # Wrong method for ilo version error
+            # uid_control
+
+            if os.environ.get('HPILO_TEST_CHANGES', None) == 'OK':
+                pass
+                # Tests that make changes:
+                # activate_license
+                # add_user/delete_user/import_ssh_key for user/delete_ssh_key
+                # mod_global_settings ssh_port 2222 / 22
+
+            def issubset(self, a, b):
+                for elt in a:
+                    if elt not in b:
+                        return False
+                return  True
+
+        self.IloTest = IloTest
+        unittest.main(self, argv=[sys.argv[0], 'IloTest'])
